/**
 *      
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2006 by Bernd Breitenbach.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 */


/**
 * \file mazeintern.h
 * \author Bernd Breitenbach
 * \remarks
 *      bb - Aug 27, 2006: Created.
 */


#ifndef _MAZEINTERN_H
#define _MAZEINTERN_H

#include "mazelib.h"

#define CELL_TYPE(c)          ((c)->type)
#define CELL_STATE(c)         ((c)->state)
#define CELL_VALID(c)         ((c)->type<TYPE_VALID)
#define CELL_NO(c)            ((c)->no)
#define CELL_X(c)             ((c)->coords.cartesic.x)
#define CELL_Y(c)             ((c)->coords.cartesic.y)
#define CELL_Z(c)             ((c)->coords.cartesic.z)
#define CELL_R(c)             ((c)->coords.polar.r)
#define CELL_A(c)             ((c)->coords.polar.a)
#define CELL_N_NEIGHBOURS(c)  ((c)->n_neighbours)
#define CELL_NEIGHBOURS(c)    ((c)->neighbours)
#define CELL_NEIGHBOUR(c,idx) ((c)->neighbours[(idx+(c)->n_neighbours)%(c)->n_neighbours].neighbour)
#define CELL_NEIGHBOUR_DATA(c,idx) ((c)->neighbours[(idx+(c)->n_neighbours)%(c)->n_neighbours].user.long_data)
#define CELL_NEIGHBOUR_IDX(c,idx) ((c)->neighbours[(idx+(c)->n_neighbours)%(c)->n_neighbours].neighbour_idx)
#define CELL_WALL(c,idx)      ((c)->neighbours[(idx+(c)->n_neighbours)%(c)->n_neighbours].wall)
#define CELL_USER_PTR(c,type) ((type *)((c)->user.ptr_data))
#define CELL_USER_DATA(c)     (((c)->user.long_data))
#define CELL_BORDER(c)        ((c)->border)
#define CELL_NEIGHBOUR_DX(c,i) ((c)->border[i].neighbour_x)
#define CELL_NEIGHBOUR_DY(c,i) ((c)->border[i].neighbour_y)
#define CELL_BORDER_X(c,i)    ((c)->border[i].point.cartesic.x)
#define CELL_BORDER_Y(c,i)    ((c)->border[i].point.cartesic.y)


#define MAZE_TYPE(l)     ((l)->type)
#define MAZE_STRUCT(l)   ((l)->structure)
#define MAZE_COLUMNS(l)  ((l)->columns)
#define MAZE_ROWS(l)     ((l)->rows)
#define MAZE_TILEABLE(l) ((l)->tileable)
#define MAZE_CELLS(l)    ((l)->cells)
#define MAZE_CELL(l,i)   (&(l)->cells[i])
#define MAZE_OFF(l)      (&(l)->cells[(l)->n_cells])
#define MAZE_N_CELLS(l)  ((l)->n_cells)
#define MAZE_ENTRANCE(l) ((l)->entrance)
#define MAZE_EXIT(l)     ((l)->exit)

#define STRUCT_NAME(s)       ((s)->name)
#define STRUCT_ROWS(s)       ((s)->rows)
#define STRUCT_COLUMNS(s)    ((s)->columns)
#define STRUCT_XMINADD(s)    ((s)->xminadd)
#define STRUCT_XMAXADD(s)    ((s)->xmaxadd)
#define STRUCT_YMINADD(s)    ((s)->yminadd)
#define STRUCT_YMAXADD(s)    ((s)->ymaxadd)
#define STRUCT_WIDTH(s)      ((s)->width)
#define STRUCT_HEIGHT(s)     ((s)->height)
#define STRUCT_CENTER_X(s,xi,yi)  ((s)->center_points[yi*(s)->columns+xi].cartesic.x)
#define STRUCT_CENTER_Y(s,xi,yi)  ((s)->center_points[yi*(s)->columns+xi].cartesic.y)
#define STRUCT_CENTER_Z(s,xi,yi)  ((s)->center_points[yi*(s)->columns+xi].cartesic.z)
#define STRUCT_CELL_TYPE(s,xi,yi) ((s)->cell_types[yi*(s)->columns+xi])
#define STRUCT_N_NEIGHBOURS(s,t)  ((s)->n_neighbours[t])
#define STRUCT_DX(s,n,t)     ((s)->cell_coords[n][t].neighbour_x)
#define STRUCT_DY(s,n,t)     ((s)->cell_coords[n][t].neighbour_y)
#define STRUCT_PX(s,n,t)     ((s)->cell_coords[n][t].point.cartesic.x)
#define STRUCT_PY(s,n,t)     ((s)->cell_coords[n][t].point.cartesic.y)
#define STRUCT_BORDER(s,n)   ((s)->cell_coords[n])


#define CARVER_TYPE(c)      ((c)->type)
#define CARVER_LABY(c)      ((c)->laby)
#define CARVER_CARVE(c)     ((c)->carve)
#define CARVER_CURRENT(c)   ((c)->current)
#define CARVER_BEGIN(r)     ((r)->begin)
#define CARVER_DIRECTION(r) ((r)->direction)
#define CARVER_ENTRANCE(r)  ((r)->entrance)
#define CARVER_EXIT(r)      ((r)->exit)
#define CARVER_PARA(r)      ((r)->para)

#define PATH_N_STEPS(p) ((p)->n_steps)
#define PATH_STEPS(p)   ((p)->steps)
#define PATH_STEP(p,n)  (&(p)->steps[n])

#define POINT_CELL(s)  ((s)->cell)
#define POINT_INDEX(s) ((s)->idx)

#define STROKE_N_POINTS(s)   ((s)->n_points)
#define STROKE_MAX_POINTS(s) ((s)->max_points)
#define STROKE_X(s,n)        ((s)->points[n].point.cartesic.x)
#define STROKE_Y(s,n)        ((s)->points[n].point.cartesic.y)
#define STROKE_TYPE(s,n)     ((s)->points[n].type)
#define STROKE_CELL(s,n)     ((s)->points[n].cell)

#define COLL_N_ITEMS(s) ((s)->n_items)
#define COLL_MAX_ITEMS(s) ((s)->max_items)
#define COLL_ITEM(s,n,type) (type)((s)->items[n])


#define DEBUG(format, ...) fprintf (stderr, "debug: "format, __VA_ARGS__)
#define ERROR(format, ...) fprintf (stderr, "error: "format, __VA_ARGS__)
#define WARN(format, ...)  fprintf (stderr, "warning: "format, __VA_ARGS__)
#define INFO(format, ...)  fprintf (stderr, "info: "format, __VA_ARGS__)

#define EPSILON     (1e-7)
#define COORD_OFF   (-9E9)
#define COORD_ZERO  (0.)

#ifndef MIN
#define MIN(a,b) ((a)<=(b)) ? (a) : (b)
#endif 

#ifndef MAX
#define MAX(a,b) ((a)>=(b)) ? (a) : (b)
#endif

#define ALLOC(type) malloc(sizeof(type))
#define CALLOC(n,type) calloc(n,sizeof(type))
#define REALLOC(ptr,n,type) realloc(ptr,n*sizeof(type))
#define VALLOC(type,n,atype) malloc(sizeof(type)+(n*sizeof(atype)))
#define VREALLOC(ptr,type,n,atype) realloc(ptr,sizeof(type)+(n*sizeof(atype)))

#define OPPOSITE_DIRECTION(dir)  ( ((dir) + HALF_CIRCLE) % FULL_CIRCLE )


#define THETA_N_CELLS_INNER(t) (((theta_maze_2d_t*)(t))->n_cells_inner)
#define THETA_RING_INCREMENT(t) (((theta_maze_2d_t*)(t))->ring_increment)
#define THETA_RINGS(t) ((t)->rows)
#define THETA_N_CELLS_RING(t,r) ( ((r)>=0 && r< THETA_RINGS(t)) ? ((r)*THETA_RING_INCREMENT(t)+THETA_N_CELLS_INNER(t)) : 0)
#define THETA_N_CELLS(t) (((THETA_RINGS(t)+1)*THETA_RINGS(t))/2+\
                           THETA_RINGS(t)*(THETA_N_CELLS_INNER(t)/THETA_RING_INCREMENT(t)-1))*THETA_RING_INCREMENT(t)

typedef struct _theta_maze_2d
{
  maze_2d_t common;
  int n_cells_inner;
  int ring_increment;
} theta_maze_2d_t;


extern cell_structure_t ortho_struct;
extern cell_structure_t hexa_struct;
extern cell_structure_t tria_struct;
extern cell_structure_t cairo_struct;
extern cell_structure_t hexa_square_struct;
extern cell_structure_t octa_square_struct;
extern cell_structure_t rhombic_struct;
extern cell_structure_t hexa_rhombic_struct;
extern cell_structure_t triangle_square_struct;
#endif /* _MAZEINTERN_H */
